Brownian class
Generate random terrain using Brownian motion.
Note that this method takes a particularly long time to run (a few seconds).
Parameters are the same as those for {@link THREE.Terrain.DiamondSquare}.
Constructors
- Brownian(Float32List g, TerrainOptions options)
Properties
- hashCode → int
-
The hash code for this object.
no setterinherited
- runtimeType → Type
-
A representation of the runtime type of the object.
no setterinherited
Methods
-
noSuchMethod(
Invocation invocation) → dynamic -
Invoked when a nonexistent method or property is accessed.
inherited
-
toString(
) → String -
A string representation of this object.
inherited
Operators
-
operator ==(
Object other) → bool -
The equality operator.
inherited